import torch
from model import softMaskedBert, biGruDetector
from torch.utils.data import DataLoader
from torch import nn
from data import myData, collect_fn
from transformers import BertModel, BertTokenizer
from config import Config


class SmbModel(nn.Module):
    def __init__(self, config):
        super(SmbModel, self).__init__()
        self.config = config

        # bert的模块引用
        self.tokenizer = BertTokenizer.from_pretrained(self.config.vocab_file)
        self.bert = BertModel.from_pretrained(self.config.bert_path)
        self.vocab_size = self.tokenizer.vocab_size
        self.embedding = self.bert.embeddings
        self.masked_e = self.embedding(torch.tensor([[self.tokenizer.mask_token_id]], dtype=torch.long)).detach()
        self.detector_model = biGruDetector(self.config.embedding_size, self.config.hidden_size)
        self.corrector_model = softMaskedBert(
            self.config,
            vocab_size=self.vocab_size,
            masked_e=self.masked_e,
            bert_encoder=self.bert.encoder)

    def forward(self, input_, masks):
        input_ids = self.embedding(input_)
        detect_prob = self.detector_model(input_ids)
        correct_prob = self.corrector_model(input_ids,
                                            detect_prob,
                                            self.bert.get_extended_attention_mask(masks, input_ids.shape,
                                                                                  self.config.device))

        return detect_prob, correct_prob
    def inference(self, sentence):
        input_id = self.tokenizer.convert_tokens_to_ids(['[CLS]'] + list(sentence) + ['[SEP]'])
        masks = [1] * len(input_id)
        input_ = torch.tensor(input_id).view(1,-1)
        masks = torch.tensor(masks).float().view(1,-1)
        input_ids = self.embedding(input_)
        detect_prob = self.detector_model(input_ids)
        correct_prob = self.corrector_model(input_ids,
                                            detect_prob,
                                            self.bert.get_extended_attention_mask(masks, input_ids.shape,
                                                                                  self.config.device))
        detect_prob = detect_prob.flatten()
        detect_prob = detect_prob[1:-1]
        correct_prob = correct_prob[1:-1]
        return detect_prob,correct_prob

if __name__ == "__main__":
    config = Config()
    # 模型初始化
    model = SmbModel(config)
    # 数据构造
    train_dataset = myData(config, 'train')
    train_data_loader = DataLoader(dataset=train_dataset, batch_size=config.batch_size,
                                   collate_fn=collect_fn, shuffle=True)
    test_dataset = myData(config, 'dev')
    test_data_loader = DataLoader(dataset=test_dataset, batch_size=config.batch_size,
                                   collate_fn=collect_fn, shuffle=True)
    # 优化器
    optimizer = torch.optim.Adam(model.parameters(), lr=config.lr)
    # 损失函数计算
    detector_criterion = nn.BCELoss()
    corrector_criterion = nn.NLLLoss()
    # 训练
    step = 0
    
    for epoch in range(config.epoch):
        # 训练
        detect_TP = 0
        P = 0
        print('epoch:', epoch)
        model.train()
        for idx, batch_data in enumerate(train_data_loader):
#             print('epoch:', epoch, '******* step:', step)
            if hasattr(torch.cuda, 'empty_cache'):
                torch.cuda.empty_cache()
            step += 1
            batch_inp_ids, batch_out_ids, batch_labels, batch_mask, intexts, outtexts = batch_data
            batch_out_ids = batch_out_ids.to(config.device)  # 选择计算设备，下同
            batch_labels = batch_labels.to(config.device)
            batch_mask = batch_mask.to(config.device)
            batch_inp_ids = batch_inp_ids.to(config.device)
            detect_prob, correct_prob = model.forward(batch_inp_ids, batch_mask)
            detector_loss = detector_criterion(detect_prob.squeeze() * batch_mask, batch_labels.float())
            corrector_loss = corrector_criterion(
                (correct_prob * batch_mask.unsqueeze(-1)).reshape(-1, correct_prob.shape[-1]),
                batch_out_ids.reshape(-1))
            loss = config.gama * corrector_loss + (1 - config.gama) * detector_loss  # 联合loss
            optimizer.zero_grad()  # 每次迭代需梯度置零
            # 这里有个小技巧，梯度不置零可以近似获得增大batch_size的效果，以减少显存不足的限制，比如每两个batch，梯度归零一次
            # 因为tf默认执行这个操作，学习torch之后才发现的
            loss.backward(retain_graph=True)
            optimizer.step()
            for sentence_id in range(batch_inp_ids.shape[0]):
                pred = detect_prob[sentence_id].flatten().argmax()
                P += 1
                if batch_labels[sentence_id][pred]==1:
                    detect_TP += 1
            if step % 200 ==0:
                print('step:',step)
                print('detector_loss:',detector_loss)
                print('corrector_loss:',corrector_loss)
                print('train detect_TP:', detect_TP)
                print('train detect_P:', P)
         # 存模型
        torch.save(model, f'checkpoints/skpt_epoch{epoch}.pt')
        model.eval()
        detect_TP = 0
        correct_TP = 0
        P = 0
        with torch.no_grad():
            for idx, batch_data in enumerate(test_data_loader):
                batch_inp_ids, batch_out_ids, batch_labels, batch_mask, intexts, outtexts = batch_data
                batch_out_ids = batch_out_ids.to(config.device)  # 选择计算设备，下同
                batch_labels = batch_labels.to(config.device)
                batch_mask = batch_mask.to(config.device)
                batch_inp_ids = batch_inp_ids.to(config.device)
               
                detect_prob, correct_prob = model.forward(batch_inp_ids, batch_mask)

                for sentence_id in range(batch_inp_ids.shape[0]):
                    pred = detect_prob[sentence_id].flatten().argmax()
                    P += 1
                    if batch_labels[sentence_id][pred]==1:
                        detect_TP += 1
            print('eval detect_TP:', detect_TP)
            print('eval detect_P:', P)
            
            




"""
@File : config.py
@Description: 参数管理器
@Author: bin.chen
@Contact: LebesgueIntigrade@163.com
@Time: 2021/8/30
@IDE: Pycharm Professional
@REFERENCE: None
"""
import os
import platform
import torch


class Config(object):
    """
    参数管理器
    """
    def __init__(self):
        if self.platform() == 'Linux':
            current_dir = '/'.join(os.path.abspath(__file__).split('/')[:-2])
        elif self.platform() == 'Windows':
            current_dir = '/'.join(os.path.abspath(__file__).split('\\\\')[:-2])
        # 获取绝对项目地址，分割符视操作系统而定
        else:
            raise 'Unknown System'

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.inference_device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.bert_path = current_dir+'model/chinese-bert-wwm-ext'
        self.vocab_file = current_dir+'model/chinese-bert-wwm-ext/vocab.txt'
        # 预训练的模型文件地址，去transformers hugface官网下载即可
        self.hidden_size = 50  # 隐层维度
        self.embedding_size = 768  # embedding维度
        self.lr = 0.0001
        self.epoch = 30
        self.batch_size = 32
        self.gama = 0.3
        self.datapath = current_dir+'/data/merge.csv'
        self.trainpath = current_dir+'/data/train1.csv'
        self.testpath = current_dir+'/data/test1.csv'
        self.devpath = current_dir+'/data/dev1.csv'
        self.checkpoints = current_dir+'/model/checkpoints'

    @staticmethod
    def platform():
        """
        Returns:
            返回当前操作系统名称,'Linux'或'Windows'
        """
        return platform.system()
























import pandas as pd
from inference import Corrector
from softmaskedbert3 import SmbModel


if __name__ == "__main__":
    # 整句的纠错结果
    # Presion: 0.7884302349847656
    # Recall: 0.8384748546318079
    # F1:0.8126
    corrector = Corrector()
    data = pd.read_csv('data/dev1.csv')
    # 记录纠错准确率
    TP, TN, FP, FN = 0, 0, 0, 0
    for i, row in data.iterrows():
        random_text = row[1]
        origin_text = row[0]
        correct_text, details = corrector.correct(random_text)
        # 判断句子是否有误
        # 1.无误
        if origin_text == random_text:
            if correct_text == origin_text:
                TN += 1
            else:
                FP += 1
        # 2.有误
        else:
            if correct_text == origin_text:
                TP += 1
            else:
                if len(details) == 0:
                    FN += 1
                else:
                    FP += 1
        if i % 100 == 0 and i != 0:
            print(TP, FN, TN, FN)
            print('Presion:', TP/(TP+FP))
            print('Recall:', TP/(TP+FN))


"""
@File : model.py
@Description: softmasked-bert文本纠错模型
@Author: bin.chen
@Contact: LebesgueIntigrade@163.com
@Time: 2021/8/30
@IDE: Pycharm Professional
@REFERENCE: 论文：《Spelling Error Correction with Soft-Masked BERT》，
            模型构建参考自https://github.com/will-wiki/softmasked-bert，
            修改了部分代码，修改了项目架构，修复了部分bug，补充了一些缺失的模型结构，在loss和准确率计算中加入了mask，并增添了海量注释。
"""

import torch
import torch.nn as nn
from transformers import BertModel, BertTokenizer
import os


class bert(nn.Module):
    """
    需要用到的bert相关组件
    """
    def __init__(self, config):
        """
        初始化
        Args:
            config: 实例化的参数管理器
        """
        super().__init__()
        self.config = config
        self.bert = BertModel.from_pretrained(self.config.bert_path)
        # 加载预训练的模型
        self.embedding = self.bert.embeddings # 实例化BertEmbeddings类
        self.bert_encoder = self.bert.encoder
        # 实例化BertEncoder类，即attention结构，默认num_hidden_layers=12，也可以去本地bert模型的config.json文件里修改
        # 论文里也是12，实际运用时有需要再改
        # 查了源码，BertModel这个类还有BertEmbeddings、BertEncoder、BertPooler属性，在此之前我想获得bert embeddings都是直接用BertModel的call方法的，学习了
        self.tokenizer = BertTokenizer.from_pretrained(self.config.vocab_file)  # 加载tokenizer
        self.masked_e = self.embedding(torch.tensor([[self.tokenizer.mask_token_id]], dtype=torch.long))
        # 加载[mask]字符对应的编码，并计算其embedding
        self.vocab_size = self.tokenizer.vocab_size  # 词汇量


class biGruDetector(nn.Module):
    """
    论文中的检测器
    """
    def __init__(self, input_size, hidden_size, num_layer=1):
        """
        类初始化
        Args:
            input_size: embedding维度
            hidden_size: gru的隐层维度
            num_layer: gru层数
        """
        super(biGruDetector, self).__init__()
        self.rnn = nn.GRU(input_size, hidden_size, num_layers=num_layer,
                          bidirectional=True, batch_first=True)
        # GRU层
        self.linear = nn.Linear(hidden_size * 2, 1)
        # 线性层
        # 因为双向GRU，所以输入维度是hidden_size*2；因为只需要输出个概率，所以第二个维度是1

    def forward(self, inp):
        """
        类call方法的覆盖
        Args:
            inp: 输入数据，embedding之后的！形如[batch_size,sequence_length,embedding_size]

        Returns:
            模型输出
        """
        rnn_output, _ = self.rnn(inp)
        # rnn输出output和最后的hidden state，这里只需要output；
        # 在batch_first设为True时，shape为（batch_size,sequence_length,2*hidden_size）;
        # 因为是双向的，所以最后一个维度是2*hidden_size。
        output = nn.Sigmoid()(self.linear(rnn_output))
        # sigmoid函数，没啥好说的，论文里就是这个结构
        return output
        # output维度是[batch_size, sequence_length, 1]


class softMaskedBert(nn.Module):
    """
    softmasked bert模型
    """
    def __init__(self, config, **kwargs):
        """
        类初始化
        Args:
            config: 实例化的参数管理器
        """
        super(softMaskedBert, self).__init__()
        self.config = config  # 加载参数管理器
        self.vocab_size = kwargs['vocab_size']
        self.masked_e = kwargs['masked_e']
        self.bert_encoder = kwargs['bert_encoder']

        self.linear = nn.Linear(self.config.embedding_size, self.vocab_size).to(config.device)  # 线性层，没啥好说的
        self.log_softmax = nn.LogSoftmax(dim=-1).to(config.device)
        # LogSoftmax就是对softmax取log

    def forward(self, bert_embedding, p, input_mask=None):
        """
        call方法
        Args:
            bert_embedding: 输入序列的bert_embedding
            p: 检测器的输出，表示输入序列对应位置的字符错误概率，维度：[batch_size, sequence_length, 1]
            input_mask: extended_attention_mask，不是单纯的输入序列的mask，具体使用方法见下面的代码注释
        Returns:
            模型输出，经过了softmax和log，维度[batch_size,sequence_length,num_vocabulary]
        """
        soft_bert_embedding = p * self.masked_e + (1 - p) * bert_embedding  # detector输出和[mask]的embedding加权求和
        bert_out = self.bert_encoder(hidden_states=soft_bert_embedding, attention_mask=input_mask)
        # 之后再经过一个encoder结构
        # 这里有个大坑，原本看transformer的手册，BertModel的attention mask是用于输入的遮挡，维度是[batch_size,sequence_length]，但是这么输入肯定报错，
        # 查源码得知，encoder使用这个参数的时候做了处理（主要是因为多头注意力机制），直接在encoder模块里传入最初的mask就会报错
        # 源代码备注：
        # Provided a padding mask of dimensions [batch_size, seq_length]
        # - if the model is a decoder, apply a causal mask in addition to the padding mask
        # - if the model is an encoder, make the mask broadcastable to [batch_size, num_heads, seq_length, seq_length]
        # 不能直接传入维度是[batch_size,sequence_length]的mask！具体处理方案见train.py
        h = bert_out[0] + bert_embedding  # 残差
        linearout = self.linear(h)
        out = self.log_softmax(linearout) # 线性层，再softmax输出
        # out维度：[batch_size,sequence_length,num_vocabulary]
        return out




import torch
from config import Config
from transformers import BertTokenizer


class Corrector(object):
    def __init__(self):
        super(Corrector, self).__init__()
        self.config = Config()
        self.tokenizer = BertTokenizer.from_pretrained(self.config.vocab_file)
        self.model = torch.load('model/checkpoints/skpt_epoch5.pt', map_location = self.config.device)
        self.model.eval()
        self.vocabs = []
        with open('model/chinese-bert-wwm-ext/vocab.txt', 'r', encoding='utf-8') as f:
            for char in f.readlines():
                self.vocabs.append(char.strip())

    def correct(self, sentence):
        input_id = self.tokenizer.convert_tokens_to_ids(['[CLS]'] + list(sentence) + ['[SEP]'])
        mask = [1] * len(input_id)
        input_id = torch.tensor(input_id).view(1,-1)
        mask = torch.tensor(mask).float().view(1,-1)
        detect_prob, correct_prob = self.model.forward(input_id.to(self.config.device), mask.to(self.config.device))
        detect_prob = detect_prob.flatten()
        detect_prob = detect_prob[1:-1]
        correct_text = sentence
        details = {}
        for idx in range(len(detect_prob)):
            if detect_prob[idx] > 0.1:
                #这里是idx+1的原因是correct的返回结果包括增加的首尾（不同于detect）
                t = correct_prob[0][idx+1].argmax()
                word = self.vocabs[t]
                correct_text = list(correct_text)
                correct_text[idx] = word
                correct_text = ''.join(correct_text)
                if sentence[idx] != word:
                    details[sentence[idx]] = word
        return correct_text, details



# coding: utf-8

import setuptools

setuptools.setup(
    name='softmaskedbert',  # 项目名称,也就是pip list后会出来的包名
    author='torchCoder',
    version='1.0.0',
    packages=setuptools.find_packages(),  # 包含所有的py文件
    include_package_data=True,  # 将数据文件也打包
    zip_safe=True
)



from inference import Corrector
from softmaskedbert3 import SmbModel


corrector = Corrector()
sentence = '她回国以后，男朋友就把他拍的像片寄给她。'
csentence = corrector.correct(sentence)
print(sentence)
print(csentence)


"""
@File : data.py
@Description: 加载数据，并进行部分预处理
@Author: bin.chen
@Contact: LebesgueIntigrade@163.com
@Time: 2021/8/31
@IDE: Pycharm Professional
@REFERENCE: 关于torch.utils.data的中文文档：
            https://pytorch-cn.readthedocs.io/zh/latest/package_references/data/

            关于collect_fn参数的小技巧：
            https://zhuanlan.zhihu.com/p/361830892

            关于迭代式数据集和映射式数据集的区别：
            https://pytorch.apachecn.org/docs/1.4/96.html
"""

import torch
import pandas as pd
from transformers import BertTokenizer
from torch.utils.data import Dataset, DataLoader
from torch.nn.utils.rnn import pad_sequence


class myData(Dataset):
    """
    数据管理器，继承自torch的Dataset类，映射式数据集
    """

    def __init__(self, config, type):
        """
        初始化
        Args:
            config: 参数管理器
        """
        super(myData, self).__init__()
        assert type in ['train', 'test', 'dev']
        if type == 'train':
            self.data = pd.read_csv(config.trainpath)  # 加载数据
        elif type == 'test':
            self.data = pd.read_csv(config.testpath)
        elif type == 'dev':
            self.data = pd.read_csv(config.devpath)
        self.tokenizer = BertTokenizer.from_pretrained(config.vocab_file)  # 默认就会添加起止符、未知符等设置

    def __len__(self):
        """
        按照官网手册，这是必须实现的方法
        Returns:
            数据集长度
        """
        return self.data.shape[0]

    def __getitem__(self, item):
        """
        也是必须实现的方法
        Args:
            item: 手册指定参数，目测是int型数字

        Returns:
            一条数据，可以自定义，比如这里我就返回了输入文本id，输出文本id，错字概率列表(标签)，mask矩阵
        """
        data = self.data.iloc[item]  # pd读指定行的数
        inp_txt = data['random_text']  # 输入文本，包含错字
        out_txt = data['origin_text']  # 输出文本，完全正确
        label = [0] + [int(x) for x in data['label'].strip().split()] + [0]
        # 列表，标志着文本对错，如[0,0,0,1]，就表示第五个字符是错的
        inp_ids = self.tokenizer.convert_tokens_to_ids(['[CLS]'] + list(inp_txt) + ['[SEP]'])
        out_ids = self.tokenizer.convert_tokens_to_ids(['[CLS]'] + list(out_txt) + ['[SEP]'])
        # 文字转编码；
        # 直接用encode函数来编码的话，有些特殊词会基于词粒度，所以这里选择先一个字一个字划分开，再用convert_tokens_to_ids函数；
        # 因为要用convert_tokens_to_ids来编码，而不是encode，所以要手动添加起止符。
        mask = [1] * len(inp_ids)
        # 遮挡
        return torch.tensor(inp_ids), torch.tensor(out_ids), torch.tensor(label).float(), torch.tensor(mask).float(), \
               list(inp_txt), list(out_txt)

        # transformers建议模型输入是torch.long，即int64，这里用torch.tensor转换类型之后，默认就是int64


def collect_fn(batch_data):
    """
    自定义的collect_fn函数，详见文件头注释中的REFERENCE部分
    Args:
        batch_data: 一个batch的数据

    Returns:
        填充并转tensor之后的数据
    """
    batch_inp_ids = [data[0] for data in batch_data]
    batch_out_ids = [data[1] for data in batch_data]
    batch_label = [data[2] for data in batch_data]
    batch_mask = [data[3] for data in batch_data]
    batch_input = [data[4] for data in batch_data]
    batch_output = [data[5] for data in batch_data]
    # 从batch中取数

    batch_inp_ids = pad_sequence(batch_inp_ids, batch_first=True)
    batch_out_ids = pad_sequence(batch_out_ids, batch_first=True)
    batch_label = pad_sequence(batch_label, batch_first=True)
    batch_mask = pad_sequence(batch_mask, batch_first=True)
    # 填充，默认以最大序列长度填充

    return batch_inp_ids, batch_out_ids, batch_label, batch_mask, batch_input, batch_output


if __name__ == '__main__':
    from config import Config

    config = Config()
    torch_dataset = myData(config)
    data_loader = DataLoader(dataset=torch_dataset, batch_size=2, collate_fn=collect_fn)
    for i in data_loader:
        print(i)
        break
